#  -*- coding: utf-8 -*-
"""Tests for sksrurgerycalibration triangulate"""

import time

import numpy as np

import sksurgerycalibration.algorithms.triangulate as at


def load_chessboard_arrays():
    """
    Load array of data for "Chessboard Test"
    """

    points_in_2d = np.zeros((4, 4), dtype=np.double)
    points_in_2d[0, 0] = 1100.16
    points_in_2d[0, 1] = 262.974
    points_in_2d[0, 2] = 1184.84
    points_in_2d[0, 3] = 241.915
    points_in_2d[1, 0] = 1757.74
    points_in_2d[1, 1] = 228.971
    points_in_2d[1, 2] = 1843.52
    points_in_2d[1, 3] = 204.083
    points_in_2d[2, 0] = 1065.44
    points_in_2d[2, 1] = 651.593
    points_in_2d[2, 2] = 1142.75
    points_in_2d[2, 3] = 632.817
    points_in_2d[3, 0] = 1788.22
    points_in_2d[3, 1] = 650.41
    points_in_2d[3, 2] = 1867.78
    points_in_2d[3, 3] = 632.59

    left_undistorted = np.zeros((4, 2), dtype=np.double)
    left_undistorted[0, 0] = points_in_2d[0, 0]
    left_undistorted[0, 1] = points_in_2d[0, 1]
    left_undistorted[1, 0] = points_in_2d[1, 0]
    left_undistorted[1, 1] = points_in_2d[1, 1]
    left_undistorted[2, 0] = points_in_2d[2, 0]
    left_undistorted[2, 1] = points_in_2d[2, 1]
    left_undistorted[3, 0] = points_in_2d[3, 0]
    left_undistorted[3, 1] = points_in_2d[3, 1]

    right_undistorted = np.zeros((4, 2), dtype=np.double)
    right_undistorted[0, 0] = points_in_2d[0, 2]
    right_undistorted[0, 1] = points_in_2d[0, 3]
    right_undistorted[1, 0] = points_in_2d[1, 2]
    right_undistorted[1, 1] = points_in_2d[1, 3]
    right_undistorted[2, 0] = points_in_2d[2, 2]
    right_undistorted[2, 1] = points_in_2d[2, 3]
    right_undistorted[3, 0] = points_in_2d[3, 2]
    right_undistorted[3, 1] = points_in_2d[3, 3]

    left_intrinsic = np.eye(3, dtype=np.double)
    left_intrinsic[0, 0] = 2012.186314
    left_intrinsic[1, 1] = 2017.966019
    left_intrinsic[0, 2] = 944.7173708
    left_intrinsic[1, 2] = 617.1093984

    right_intrinsic = np.eye(3, dtype=np.double)
    right_intrinsic[0, 0] = 2037.233928
    right_intrinsic[1, 1] = 2052.018948
    right_intrinsic[0, 2] = 1051.112809
    right_intrinsic[1, 2] = 548.0675962

    left_to_right_rotation = np.eye(3, dtype=np.double)
    left_to_right_rotation[0, 0] = 0.999678
    left_to_right_rotation[0, 1] = 0.000151
    left_to_right_rotation[0, 2] = 0.025398
    left_to_right_rotation[1, 0] = -0.000720
    left_to_right_rotation[1, 1] = 0.999749
    left_to_right_rotation[1, 2] = 0.022394
    left_to_right_rotation[2, 0] = -0.025388
    left_to_right_rotation[2, 1] = -0.022405
    left_to_right_rotation[2, 2] = 0.999426

    left_to_right_translation = np.zeros((3, 1), dtype=np.double)
    left_to_right_translation[0, 0] = -4.631472
    left_to_right_translation[1, 0] = 0.268695
    left_to_right_translation[2, 0] = 1.300256

    model_points = np.zeros((4, 3), dtype=np.double)
    model_points[0, 0] = 0
    model_points[0, 1] = 0
    model_points[0, 2] = 0
    model_points[1, 0] = 39
    model_points[1, 1] = 0
    model_points[1, 2] = 0
    model_points[2, 0] = 0
    model_points[2, 1] = 27
    model_points[2, 2] = 0
    model_points[3, 0] = 39
    model_points[3, 1] = 27
    model_points[3, 2] = 0

    left_rotation = np.zeros((3, 3), dtype=np.double)
    left_rotation[0, 0] = 0.966285949
    left_rotation[0, 1] = -0.1053020017
    left_rotation[0, 2] = 0.2349530874
    left_rotation[1, 0] = -0.005105986897
    left_rotation[1, 1] = 0.9045241988
    left_rotation[1, 2] = 0.4263917244
    left_rotation[2, 0] = -0.2574206552
    left_rotation[2, 1] = -0.4132159994
    left_rotation[2, 2] = 0.8734913532

    left_translation = np.zeros((3, 1), dtype=np.double)
    left_translation[0, 0] = 9.847672184
    left_translation[1, 0] = -22.45992103
    left_translation[2, 0] = 127.7836183

    return points_in_2d, \
           left_undistorted, \
           right_undistorted, \
           left_intrinsic, \
           right_intrinsic, \
           left_to_right_rotation, \
           left_to_right_translation, \
           model_points, \
           left_rotation, \
           left_translation


def rms_between_points(a_array, b_array):
    """
    Compute Root Mean Squater between_corresponding_points a, b.
    """
    rms = 0
    diff = 0
    squared_diff = 0

    if a_array.shape[0] != b_array.shape[0]:
        print(f' a has {a_array.shape[0]} rows but b has {b_array.shape[0]} rows')

    if a_array.shape[1] != 3:
        print(f'a does not have 3 columns but {a_array.shape[1]} columns')

    if b_array.shape[1] != 3:
        print(f'b does not have 3 columns but {b_array.shape[1]} columns')

    for dummy_row_index in range(a_array.shape[0]):
        for dummy_col_index in range(a_array.shape[1]):
            diff = b_array[dummy_row_index, dummy_col_index] - a_array[dummy_row_index, dummy_col_index]
            squared_diff = diff * diff
            rms += squared_diff

    rms /= a_array.shape[0]
    rms = np.sqrt(rms)
    return rms


def test_triangulate_points_hartley():
    """
    Test triangulate points with hartley using "Chessboard Test"
    """

    points_in_2d, _left_undistorted, _right_undistorted, left_intrinsic, right_intrinsic, \
    left_to_right_rotation, left_to_right_translation, model_points, left_rotation, left_translation = load_chessboard_arrays()

    model_points_transposed = model_points.T
    rotated_model_points = np.zeros((model_points_transposed.shape[0], model_points_transposed.shape[1]),
                                    dtype=np.double)
    # rotated_model_points = cv2.gemm(src1=left_rotation, src2=model_points_transposed, alpha=1.0, src3=None,
    #                                beta=0.0)  # flags=cv2.GEMM_2_T?
    rotated_model_points = left_rotation.dot(model_points_transposed)
    model_points_rotated_transposed = rotated_model_points.T
    transformed_model_points = np.zeros(
        (model_points_rotated_transposed.shape[0], model_points_rotated_transposed.shape[1]), dtype=np.double)

    for dummy_row_index in range(model_points_rotated_transposed.shape[0]):
        for dummy_col_index in range(model_points_rotated_transposed.shape[1]):
            transformed_model_points[dummy_row_index, dummy_col_index] = model_points_rotated_transposed[
                                                                             dummy_row_index, dummy_col_index] + \
                                                                         left_translation[dummy_col_index, 0]

    start = time.time_ns()
    points_from_hartley = at.triangulate_points_hartley(points_in_2d,
                                                        left_intrinsic,
                                                        right_intrinsic,
                                                        left_to_right_rotation,
                                                        left_to_right_translation)
    print(f'\n Elapsed time for at.triangulate_points_hartley(): {(time.time_ns() - start) / 1e6} millisecs ')

    start = time.time_ns()
    points_from_hartley_opencv = at.triangulate_points_opencv(points_in_2d,
                                                              left_intrinsic,
                                                              right_intrinsic,
                                                              left_to_right_rotation,
                                                              left_to_right_translation)
    print(f'\n Elapsed time for at.triangulate_points_opencv(): {(time.time_ns() - start) / 1e6} millisecs')

    rms_hartley = rms_between_points(transformed_model_points, points_from_hartley)
    rms_hartley_opencv = rms_between_points(transformed_model_points, points_from_hartley_opencv)

    print(f'\n rms_hartley: {rms_hartley} and test (rms_hartley < 1.5)')
    print(f'\n rms_hartley_opencv:  {rms_hartley_opencv} and test (rms_hartley < 1.5) \n')
    assert rms_hartley < 1.5 and rms_hartley_opencv < 1.5



# -*- coding: utf-8 -*-

""" Class to do stateful video calibration of a mono camera. """

import copy
import logging

import sksurgeryimage.calibration.point_detector as pd

import sksurgerycalibration.video.video_calibration_data as cd
import sksurgerycalibration.video.video_calibration_driver_base as vdb
import sksurgerycalibration.video.video_calibration_params as cp
import sksurgerycalibration.video.video_calibration_utils as cu
import sksurgerycalibration.video.video_calibration_wrapper as vc

LOGGER = logging.getLogger(__name__)


class MonoVideoCalibrationDriver(vdb.BaseVideoCalibrationDriver):
    """ Class to do stateful video calibration of a mono camera. """
    def __init__(self,
                 point_detector: pd.PointDetector,
                 minimum_points_per_frame: int
                 ):
        """
        Stateful class for mono video calibration.

        :param point_detector: Class derived from PointDetector
        :param minimum_points_per_frame: Minimum number to accept frame
        """
        super().__init__(minimum_points_per_frame)

        self.point_detector = point_detector

        # Create data holders, and parameter holders, specific to Mono.
        calibration_data = cd.MonoVideoData()
        calibration_params = cp.MonoCalibrationParams()

        # Pass them to base class, so base class can access them.
        self._init_internal(calibration_data, calibration_params)

    def grab_data(self,
                  image,
                  device_tracking=None,
                  calibration_object_tracking=None):
        """
        Extracts points, by passing it to the PointDetector.

        This will throw various exceptions if the input data is invalid,
        but will return empty arrays if no points were detected.
        So, no points is not an error. Its an expected condition.

        :param image: RGB image.
        :param device_tracking: transformation for the tracked device
        :param calibration_object_tracking: transformation of tracked
        calibration object
        :return: The number of points grabbed.
        """
        number_of_points = 0

        ids, object_points, image_points = \
            self.point_detector.get_points(image)

        if ids is not None and ids.shape[0] >= self.minimum_points_per_frame:

            ids, image_points, object_points = \
                cu.convert_pd_to_opencv(ids,
                                        object_points,
                                        image_points)

            self.video_data.push(image,
                                 ids,
                                 object_points,
                                 image_points)

            self.tracking_data.push(device_tracking,
                                    calibration_object_tracking)

            number_of_points = image_points.shape[0]

        LOGGER.info("Grabbed: Returning %s points.", str(number_of_points))
        return number_of_points

    def calibrate(self, flags=0):
        """
        Do the video calibration, returning RMS re-projection error.

        :param flags: OpenCV calibration flags, eg. cv2.CALIB_FIX_ASPECT_RATIO
        :return: RMS projection
        """
        rms_proj_err, camera_matrix, dist_coeffs, rvecs, tvecs = \
            vc.mono_video_calibration(
                self.video_data.object_points_arrays,
                self.video_data.image_points_arrays,
                (self.video_data.images_array[0].shape[1],
                 self.video_data.images_array[0].shape[0]),
                flags
            )

        self.calibration_params.set_data(camera_matrix,
                                         dist_coeffs,
                                         rvecs,
                                         tvecs)

        LOGGER.info("Mono calibration: rms_proj_err=%s.", str(rms_proj_err))
        return rms_proj_err, copy.deepcopy(self.calibration_params)

    def iterative_calibration(self,
                              number_of_iterations: int,
                              reference_ids,
                              reference_image_points,
                              reference_image_size,
                              flags: int = 0):
        """
        Does iterative calibration, like Datta 2009,
        returning RMS re-projection error.
        :return: RMS projection
        """
        rms_proj_err, param_copy = self.calibrate(flags=flags)
        cached_images = copy.deepcopy(self.video_data.images_array)

        for i in range(0, number_of_iterations):
            images = copy.deepcopy(cached_images)
            cu.detect_points_in_canonical_space(
                self.point_detector,
                self.minimum_points_per_frame,
                self.video_data,
                images,
                self.calibration_params.camera_matrix,
                self.calibration_params.dist_coeffs,
                reference_ids,
                reference_image_points,
                reference_image_size)

            rms_proj_err, param_copy = self.calibrate(flags=flags)

            self.point_detector.set_camera_parameters(
                self.calibration_params.camera_matrix,
                self.calibration_params.dist_coeffs)

            LOGGER.info("Iterative calibration: %s: rms_proj_err=%s.",
                        str(i), str(rms_proj_err))

        return rms_proj_err, param_copy

    def handeye_calibration(self,
                            override_pattern2marker=None,
                            use_opencv: bool = True,
                            do_bundle_adjust: bool = False):
        """
        Do handeye calibration, returning RMS re-projection error.

        Note: This handeye_calibration on this class assumes you are
        tracking both the calibration pattern (e.g. chessboard) and the
        device (e.g. laparoscope). So, the calibration routines calibrate
        for hand2eye and pattern2marker. If you want something more customised,
        work with video_calibration_hand_eye.py.

        :param override_pattern2marker: If provided a 4x4 pattern2marker
        that is taken as constant.
        :param use_opencv: If True we use OpenCV based methods, if false,
        Guofang Xiao's method.
        :param do_bundle_adjust: If True we do an additional bundle adjustment
        at the end.

        :return: RMS reprojection error
        :rtype: float
        """

        rms_proj_err, handeye, pattern2marker = \
            vc.mono_handeye_calibration(
                self.video_data.object_points_arrays,
                self.video_data.image_points_arrays,
                self.calibration_params.camera_matrix,
                self.calibration_params.dist_coeffs,
                self.tracking_data.device_tracking_array,
                self.tracking_data.calibration_tracking_array,
                self.calibration_params.rvecs,
                self.calibration_params.tvecs,
                override_pattern2marker=override_pattern2marker,
                use_opencv=use_opencv,
                do_bundle_adjust=do_bundle_adjust
            )

        self.calibration_params.set_handeye(handeye, pattern2marker)

        return rms_proj_err, copy.deepcopy(self.calibration_params)



# -*- coding: utf-8 -*-

""" Cost functions for video calibration, used with scipy. """

# pylint:disable=invalid-name

import numpy as np

import sksurgerycalibration.video.video_calibration_metrics as vm
import sksurgerycalibration.video.video_calibration_utils as vu


def stereo_2d_error_for_extrinsics(x_0,
                                   common_object_points,
                                   common_left_image_points,
                                   common_right_image_points,
                                   left_intrinsics,
                                   left_distortion,
                                   right_intrinsics,
                                   right_distortion,
                                   l2r_rmat,
                                   l2r_tvec
                                   ):
    """
    Computes a vector of residuals between projected image points
    and actual image points, for left and right image. x_0 should
    contain left camera extrinsic parameters.
    """
    rvecs = []
    tvecs = []
    number_of_frames = len(common_object_points)
    for i in range(0, number_of_frames):
        rvec = np.zeros((3, 1))
        rvec[0][0] = x_0[6 * i + 0]
        rvec[1][0] = x_0[6 * i + 1]
        rvec[2][0] = x_0[6 * i + 2]
        tvec = np.zeros((3, 1))
        tvec[0][0] = x_0[6 * i + 3]
        tvec[1][0] = x_0[6 * i + 4]
        tvec[2][0] = x_0[6 * i + 5]
        rvecs.append(rvec)
        tvecs.append(tvec)

    residual = vm.compute_stereo_2d_err(l2r_rmat,
                                        l2r_tvec,
                                        common_object_points,
                                        common_left_image_points,
                                        left_intrinsics,
                                        left_distortion,
                                        common_object_points,
                                        common_right_image_points,
                                        right_intrinsics,
                                        right_distortion,
                                        rvecs,
                                        tvecs,
                                        return_residuals=True
                                        )
    return residual


def mono_proj_err_h2e(x_0,
                      object_points,
                      image_points,
                      intrinsics,
                      distortion,
                      pattern_tracking,
                      device_tracking,
                      pattern2marker_matrix
                      ):
    """
    Computes the SSE of projected
    image points and actual image points, for a single camera,
    where we have a tracked calibration pattern, and assume the
    pattern2marker transform should remain fixed. Therefore we
    only optimise hand-eye. So, x_0 should be of length 6.
    """
    assert len(x_0) == 6

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i]) @ \
              pattern_tracking[i] @ pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion,
                                     return_residuals=False)
    return proj


def mono_proj_err_p2m_h2e(x_0,
                          object_points,
                          image_points,
                          intrinsics,
                          distortion,
                          pattern_tracking,
                          device_tracking
                          ):
    """
    Computes the SSE between projected
    image points to actual image points, for a single camera,
    where we have a tracked pattern. Both the
    pattern2marker and hand2eye are optimised.
    So, x_0 should be of length 12.
    """
    assert len(x_0) == 12

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    p2m = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    rvec[0][0] = x_0[6]
    rvec[1][0] = x_0[7]
    rvec[2][0] = x_0[8]

    tvec[0][0] = x_0[9]
    tvec[1][0] = x_0[10]
    tvec[2][0] = x_0[11]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i])\
              @ pattern_tracking[i] @ p2m

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion,
                                     return_residuals=False)
    return proj


def mono_proj_err_h2e_g2w(x_0,
                          object_points,
                          image_points,
                          intrinsics,
                          distortion,
                          device_tracking
                          ):
    """
    Method to the SSE of projected
    image points to actual image points, for a single camera,
    where we have an untracked pattern. Both the
    hand2eye and grid2world are optimised.
    So, x_0 should be of length 12.
    """
    assert len(x_0) == 12

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    rvec[0][0] = x_0[6]
    rvec[1][0] = x_0[7]
    rvec[2][0] = x_0[8]

    tvec[0][0] = x_0[9]
    tvec[1][0] = x_0[10]
    tvec[2][0] = x_0[11]

    g2w = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i]) @ g2w

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion,
                                     return_residuals=False)
    return proj


def mono_proj_err_h2e_int_dist(x_0,
                               object_points,
                               image_points,
                               device_tracking,
                               pattern_tracking,
                               pattern2marker_matrix
                               ):
    """
    Computes the SSE between projected
    image points to actual image points, for a single camera,
    where we have a tracked pattern. The handeye, intrinsics and
    distortion parameters are optimised.
    So, x_0 should be of length 6+4+5 = 15.
    """
    assert len(x_0) == 15

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    intrinsics = np.zeros((3, 3))
    intrinsics[0][0] = x_0[6]
    intrinsics[1][1] = x_0[7]
    intrinsics[0][2] = x_0[8]
    intrinsics[1][2] = x_0[9]

    distortion = np.zeros((1, 5))
    distortion[0][0] = x_0[10]
    distortion[0][1] = x_0[11]
    distortion[0][2] = x_0[12]
    distortion[0][3] = x_0[13]
    distortion[0][4] = x_0[14]

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i]) @ \
              pattern_tracking[i] @ pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion)
    return proj


# pylint:disable=too-many-arguments
def stereo_proj_err_h2e(x_0,
                        common_object_points,
                        common_left_image_points,
                        common_right_image_points,
                        left_intrinsics,
                        left_distortion,
                        right_intrinsics,
                        right_distortion,
                        l2r_rmat,
                        l2r_tvec,
                        device_tracking_array,
                        pattern_tracking_array,
                        left_pattern2marker_matrix=None
                        ):
    """
    Computes the SSE of projected image points
    and actual image points for left and right cameras. x_0 should contain
    the 6DOF of hand-to-eye, and if left_pattern2marker_matrix is None,
    then an additional 6DOF of pattern-to-marker. So, x_0 can be either
    length 6 or length 12.

    :param x_0:
    :param common_object_points:
    :param common_left_image_points:
    :param common_right_image_points:
    :param left_intrinsics:
    :param left_distortion:
    :param right_intrinsics:
    :param right_distortion:
    :param l2r_rmat:
    :param l2r_tvec:
    :param device_tracking_array:
    :param pattern_tracking_array:
    :param left_pattern2marker_matrix:
    :return: matrix of residuals for Levenberg-Marquardt optimisation.
    """
    rvecs = []
    tvecs = []
    number_of_frames = len(common_object_points)

    h2e_rvec = np.zeros((3, 1))
    h2e_rvec[0][0] = x_0[0]
    h2e_rvec[1][0] = x_0[1]
    h2e_rvec[2][0] = x_0[2]

    h2e_tvec = np.zeros((3, 1))
    h2e_tvec[0][0] = x_0[3]
    h2e_tvec[1][0] = x_0[4]
    h2e_tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(h2e_rvec, h2e_tvec)

    if left_pattern2marker_matrix is None:

        p2m_rvec = np.zeros((3, 1))
        p2m_rvec[0][0] = x_0[6]
        p2m_rvec[1][0] = x_0[7]
        p2m_rvec[2][0] = x_0[8]

        p2m_tvec = np.zeros((3, 1))
        p2m_tvec[0][0] = x_0[9]
        p2m_tvec[1][0] = x_0[10]
        p2m_tvec[2][0] = x_0[11]

        p2m = vu.extrinsic_vecs_to_matrix(p2m_rvec, p2m_tvec)

    else:

        p2m = left_pattern2marker_matrix

    for i in range(0, number_of_frames):

        p2c = h2e \
              @ np.linalg.inv(device_tracking_array[i]) \
              @ pattern_tracking_array[i] \
              @ p2m

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_stereo_2d_err(l2r_rmat,
                                       l2r_tvec,
                                       common_object_points,
                                       common_left_image_points,
                                       left_intrinsics,
                                       left_distortion,
                                       common_object_points,
                                       common_right_image_points,
                                       right_intrinsics,
                                       right_distortion,
                                       rvecs,
                                       tvecs
                                       )
    return proj


def stereo_proj_err_h2e_int_dist_l2r(x_0,
                                     common_object_points,
                                     common_left_image_points,
                                     common_right_image_points,
                                     device_tracking_array,
                                     pattern_tracking_array,
                                     left_pattern2marker_matrix
                                     ):
    """
    Computes the SSE of projected image points against actual
    image points. x_0 should be 30 DOF.
    """
    h2e_rvec = np.zeros((3, 1))
    h2e_rvec[0][0] = x_0[0]
    h2e_rvec[1][0] = x_0[1]
    h2e_rvec[2][0] = x_0[2]

    h2e_tvec = np.zeros((3, 1))
    h2e_tvec[0][0] = x_0[3]
    h2e_tvec[1][0] = x_0[4]
    h2e_tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(h2e_rvec, h2e_tvec)

    l2r_rvec = np.zeros((3, 1))
    l2r_rvec[0][0] = x_0[6]
    l2r_rvec[1][0] = x_0[7]
    l2r_rvec[2][0] = x_0[8]

    l2r_tvec = np.zeros((3, 1))
    l2r_tvec[0][0] = x_0[9]
    l2r_tvec[1][0] = x_0[10]
    l2r_tvec[2][0] = x_0[11]

    l2r = vu.extrinsic_vecs_to_matrix(l2r_rvec, l2r_tvec)

    left_intrinsics = np.zeros((3, 3))
    left_intrinsics[0][0] = x_0[12]
    left_intrinsics[1][1] = x_0[13]
    left_intrinsics[0][2] = x_0[14]
    left_intrinsics[1][2] = x_0[15]

    left_distortion = np.zeros((1, 5))
    left_distortion[0][0] = x_0[16]
    left_distortion[0][1] = x_0[17]
    left_distortion[0][2] = x_0[18]
    left_distortion[0][3] = x_0[19]
    left_distortion[0][4] = x_0[20]

    right_intrinsics = np.zeros((3, 3))
    right_intrinsics[0][0] = x_0[21]
    right_intrinsics[1][1] = x_0[22]
    right_intrinsics[0][2] = x_0[23]
    right_intrinsics[1][2] = x_0[24]

    right_distortion = np.zeros((1, 5))
    right_distortion[0][0] = x_0[25]
    right_distortion[0][1] = x_0[26]
    right_distortion[0][2] = x_0[27]
    right_distortion[0][3] = x_0[28]
    right_distortion[0][4] = x_0[29]

    rvecs = []
    tvecs = []
    number_of_frames = len(common_object_points)

    for i in range(0, number_of_frames):

        p2c = h2e \
              @ np.linalg.inv(device_tracking_array[i]) \
              @ pattern_tracking_array[i] \
              @ left_pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_stereo_2d_err(l2r[0:3, 0:3],
                                       l2r[0:3, 3],
                                       common_object_points,
                                       common_left_image_points,
                                       left_intrinsics,
                                       left_distortion,
                                       common_object_points,
                                       common_right_image_points,
                                       right_intrinsics,
                                       right_distortion,
                                       rvecs,
                                       tvecs
                                       )
    return proj



# -*- coding: utf-8 -*-

""" Utility functions to load data. """

import os
from typing import Tuple
import numpy as np
import cv2


def get_calib_data(directory: str, idx: int) -> \
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """ Generate filenames for calibration data in a given directory. """
    left_image = cv2.imread(
        os.path.join(directory, f'calib.left.images.{idx}.png')
    )

    right_image = cv2.imread(
        os.path.join(directory, f'calib.right.images.{idx}.png')
    )

    chessboad_tracking_file = \
        os.path.join(directory,
                     f'calib.calib_obj_tracking.{idx}.txt')

    if os.path.isfile(chessboad_tracking_file):
        chessboard_tracking = np.loadtxt(chessboad_tracking_file)
    else:
        chessboard_tracking = None

    scope_tracking = np.loadtxt(
        os.path.join(directory, f'calib.device_tracking.{idx}.txt')
    )

    return left_image, right_image, chessboard_tracking, scope_tracking



# -*- coding: utf-8 -*-

""" Class to do stateful video calibration of a stereo camera. """

import copy
import logging

import cv2
import sksurgeryimage.calibration.point_detector as pd

import sksurgerycalibration.video.video_calibration_data as cd
import sksurgerycalibration.video.video_calibration_driver_base as vdb
import sksurgerycalibration.video.video_calibration_params as cp
import sksurgerycalibration.video.video_calibration_utils as cu
import sksurgerycalibration.video.video_calibration_wrapper as vw

LOGGER = logging.getLogger(__name__)


class StereoVideoCalibrationDriver(vdb.BaseVideoCalibrationDriver):
    """
    Class to do stateful video calibration of a stereo camera.
    """
    def __init__(self,
                 left_point_detector: pd.PointDetector,
                 right_point_detector: pd.PointDetector,
                 minimum_points_per_frame: int
                 ):
        """
        Stateful class for stereo video calibration.

        :param left_point_detector: Class derived from PointDetector
        :param right_point_detector: Class derived from PointDetector
        :param minimum_points_per_frame: Minimum number to accept frame
        """
        super().__init__(minimum_points_per_frame)

        self.left_point_detector = left_point_detector
        self.right_point_detector = right_point_detector

        # Create data holders, and parameter holders, specific to Stereo.
        calibration_data = cd.StereoVideoData()
        calibration_params = cp.StereoCalibrationParams()

        # Pass them to base class, so base class can access them.
        self._init_internal(calibration_data, calibration_params)

    def grab_data(self,
                  left_image,
                  right_image,
                  device_tracking=None,
                  calibration_object_tracking=None
                  ):
        """
        Extracts points, by passing it to the PointDetector.

        This will throw various exceptions if the input data is invalid,
        but will return empty arrays if no points were detected.
        So, no points is not an error. Its an expected condition.

        :param left_image: BGR image.
        :param right_image: BGR image.
        :param device_tracking: transformation for the tracked device
        :param calibration_object_tracking: transformation of tracked
        calibration object
        :return: The number of points grabbed.
        """
        number_left = 0
        number_right = 0

        # This can return None's if none are found.
        left_ids, left_object_points, left_image_points = \
            self.left_point_detector.get_points(left_image)

        if left_ids is not None:
            number_left = left_ids.shape[0]

        if number_left >= self.minimum_points_per_frame:

            right_ids, right_object_points, right_image_points = \
                self.right_point_detector.get_points(right_image)

            if right_ids is not None:
                number_right = right_ids.shape[0]

            if number_right >= self.minimum_points_per_frame:

                left_ids, left_image_points, left_object_points = \
                    cu.convert_pd_to_opencv(left_ids,
                                            left_object_points,
                                            left_image_points)

                right_ids, right_image_points, right_object_points = \
                    cu.convert_pd_to_opencv(right_ids,
                                            right_object_points,
                                            right_image_points)

                self.video_data.push(left_image,
                                     left_ids,
                                     left_object_points,
                                     left_image_points,
                                     right_image,
                                     right_ids,
                                     right_object_points,
                                     right_image_points)

                self.tracking_data.push(device_tracking,
                                        calibration_object_tracking)

        number_of_points = number_left + number_right

        LOGGER.info("Grabbed: Returning (%s+%s)=%s points.",
                    str(number_left),
                    str(number_right),
                    str(number_of_points))

        return number_left, number_right

    def calibrate(self,
                  flags=cv2.CALIB_USE_INTRINSIC_GUESS,
                  override_left_intrinsics=None,
                  override_left_distortion=None,
                  override_right_intrinsics=None,
                  override_right_distortion=None,
                  override_l2r_rmat=None,
                  override_l2r_tvec=None
                  ):
        """
        Do the stereo video calibration, returning reprojection and
        reconstruction error.

        This returns RMS projection error, which is a common metric, but also,
        the reconstruction / triangulation error.

        :param flags: OpenCV flags, eg. cv2.CALIB_FIX_INTRINSIC
        :param override_left_intrinsics:
        :param override_left_distortion:
        :param override_right_intrinsics:
        :param override_right_distortion:
        :param override_l2r_rmat:
        :param override_l2r_tvec:
        :return: projection, reconstruction error.
        :rtype: float, float
        """
        s_reproj, s_recon, \
            l_c, l_d, l_rvecs, l_tvecs, \
            r_c, r_d, r_rvecs, r_tvecs, \
            l2r_r, l2r_t, \
            essential, fundamental \
            = vw.stereo_video_calibration(
                self.video_data.left_data.ids_arrays,
                self.video_data.left_data.object_points_arrays,
                self.video_data.left_data.image_points_arrays,
                self.video_data.right_data.ids_arrays,
                self.video_data.right_data.object_points_arrays,
                self.video_data.right_data.image_points_arrays,
                (self.video_data.left_data.images_array[0].shape[1],
                 self.video_data.left_data.images_array[0].shape[0]),
                flags,
                override_left_intrinsics,
                override_left_distortion,
                override_right_intrinsics,
                override_right_distortion,
                override_l2r_rmat,
                override_l2r_tvec
            )

        self.calibration_params.set_data(l_c, l_d, l_rvecs, l_tvecs, r_c, r_d,
                                         r_rvecs, r_tvecs, l2r_r, l2r_t,
                                         essential, fundamental)

        LOGGER.info("Calibrated: proj_err=%s, recon_err=%s.",
                    str(s_reproj), str(s_recon))
        return s_reproj, s_recon, copy.deepcopy(self.calibration_params)

    # pylint:disable=too-many-arguments
    def iterative_calibration(self,
                              number_of_iterations: int,
                              reference_ids,
                              reference_image_points,
                              reference_image_size,
                              flags: int = cv2.CALIB_USE_INTRINSIC_GUESS
                              ):
        """
        Does iterative calibration, like Datta 2009, returning reprojection
        and reconstruction error.

        :return: projection, reconstruction error.
        :rtype: float, float
        """
        proj_err, recon_err, param_copy = self.calibrate(flags=flags)

        cached_left_images = copy.deepcopy(
            self.video_data.left_data.images_array)
        cached_right_images = copy.deepcopy(
            self.video_data.right_data.images_array)

        for i in range(0, number_of_iterations):

            left_images = copy.deepcopy(cached_left_images)
            right_images = copy.deepcopy(cached_right_images)

            cu.detect_points_in_stereo_canonical_space(
                self.left_point_detector,
                self.right_point_detector,
                self.minimum_points_per_frame,
                self.video_data.left_data,
                left_images,
                self.calibration_params.left_params.camera_matrix,
                self.calibration_params.left_params.dist_coeffs,
                self.video_data.right_data,
                right_images,
                self.calibration_params.right_params.camera_matrix,
                self.calibration_params.right_params.dist_coeffs,
                reference_ids,
                reference_image_points,
                reference_image_size)

            proj_err, recon_err, param_copy = \
                self.calibrate(flags)

            self.left_point_detector.set_camera_parameters(
                self.calibration_params.left_params.camera_matrix,
                self.calibration_params.left_params.dist_coeffs)

            self.right_point_detector.set_camera_parameters(
                self.calibration_params.right_params.camera_matrix,
                self.calibration_params.right_params.dist_coeffs)

            LOGGER.info("Iterative calibration: %s: proj_err=%s, recon_err=%s.",
                        str(i), str(proj_err), str(recon_err))

        return proj_err, recon_err, param_copy

    def handeye_calibration(self,
                            override_pattern2marker=None,
                            use_opencv: bool = True,
                            do_bundle_adjust: bool = False):
        """
        Do handeye calibration, returning reprojection and reconstruction error.

        Note: This handeye_calibration on this class assumes you are tracking
        both the calibration pattern (e.g. chessboard) and the device
        (e.g. laparoscope). So, the calibration routines calibrate for
        hand2eye and pattern2marker. If you want something more customised,
        work with video_calibration_hand_eye.py.

        :param override_pattern2marker: If provided a 4x4 pattern2marker
        that is taken as constant.
        :param use_opencv: If True we use OpenCV based methods, if false,
        Guofang Xiao's method.
        :param do_bundle_adjust: If True we do an additional bundle adjustment
        at the end.

        :return: reprojection, reconstruction error, camera parameters
        :rtype: float, float, object
        """

        proj_err, recon_err, l_handeye, l_pattern2marker, \
            r_handeye, r_pattern2marker = \
                vw.stereo_handeye_calibration(
                    self.calibration_params.l2r_rmat,
                    self.calibration_params.l2r_tvec,
                    self.video_data.left_data.ids_arrays,
                    self.video_data.left_data.object_points_arrays,
                    self.video_data.left_data.image_points_arrays,
                    self.video_data.right_data.ids_arrays,
                    self.video_data.right_data.image_points_arrays,
                    self.calibration_params.left_params.camera_matrix,
                    self.calibration_params.left_params.dist_coeffs,
                    self.calibration_params.right_params.camera_matrix,
                    self.calibration_params.right_params.dist_coeffs,
                    self.tracking_data.device_tracking_array,
                    self.tracking_data.calibration_tracking_array,
                    self.calibration_params.left_params.rvecs,
                    self.calibration_params.left_params.tvecs,
                    override_pattern2marker=override_pattern2marker,
                    use_opencv=use_opencv,
                    do_bundle_adjust=do_bundle_adjust
                    )

        self.calibration_params.left_params.set_handeye(
            l_handeye, l_pattern2marker)

        self.calibration_params.right_params.set_handeye(
            r_handeye, r_pattern2marker)

        return proj_err, recon_err, copy.deepcopy(self.calibration_params)



#  -*- coding: utf-8 -*-

""" Functions for Triangulation using Harley method """

import cv2
# import time
import numpy as np
import sksurgerycore.transforms.matrix as stm


def _triangulate_point_using_svd(p1_array,
                                 p2_array,
                                 u1_array,
                                 u2_array,
                                 w1_const,
                                 w2_const):
    """
    Function for Internal Triangulate Point Using SVD

    (u1_array, p1_array) is the reference pair containing normalized image coordinates (x, y) and the corresponding camera matrix.
    (u2_array, p2_array) is the second pair.

    :param p1_array: [3x4] ndarray
    :param p2_array: [3x4] ndarray
    :param u1_array: [3x1] ndarray
    :param u2_array: [3x1] ndarray
    :param w1_const: constant value
    :param w2_const: constant value

    :return x_array:  [4x1] ndarray
    """

    # Build matrix A for homogeneous equation system Ax = 0
    # Assume X = (x,y,z,1), for Linear-LS method
    # Which turns it into a AX = B system, where A is 4x3, X is 3x1 and B is 4x1
    a_array = np.zeros((4, 3), dtype=np.double)
    a_array[0, 0] = (u1_array[0][0] * p1_array[2, 0] - p1_array[0, 0]) / w1_const
    a_array[0, 1] = (u1_array[0][0] * p1_array[2, 1] - p1_array[0, 1]) / w1_const
    a_array[0, 2] = (u1_array[0][0] * p1_array[2, 2] - p1_array[0, 2]) / w1_const
    a_array[1, 0] = (u1_array[1][0] * p1_array[2, 0] - p1_array[1, 0]) / w1_const
    a_array[1, 1] = (u1_array[1][0] * p1_array[2, 1] - p1_array[1, 1]) / w1_const
    a_array[1, 2] = (u1_array[1][0] * p1_array[2, 2] - p1_array[1, 2]) / w1_const
    a_array[2, 0] = (u2_array[0][0] * p2_array[2, 0] - p2_array[0, 0]) / w2_const
    a_array[2, 1] = (u2_array[0][0] * p2_array[2, 1] - p2_array[0, 1]) / w2_const
    a_array[2, 2] = (u2_array[0][0] * p2_array[2, 2] - p2_array[0, 2]) / w2_const
    a_array[3, 0] = (u2_array[1][0] * p2_array[2, 0] - p2_array[1, 0]) / w2_const
    a_array[3, 1] = (u2_array[1][0] * p2_array[2, 1] - p2_array[1, 1]) / w2_const
    a_array[3, 2] = (u2_array[1][0] * p2_array[2, 2] - p2_array[1, 2]) / w2_const

    b_array = np.zeros((4, 1), dtype=np.double)
    b_array[0][0] = -(u1_array[0][0] * p1_array[2, 3] - p1_array[0, 3]) / w1_const
    b_array[1][0] = -(u1_array[1][0] * p1_array[2, 3] - p1_array[1, 3]) / w1_const
    b_array[2][0] = -(u2_array[0][0] * p2_array[2, 3] - p2_array[0, 3]) / w2_const
    b_array[3][0] = -(u2_array[1][0] * p2_array[2, 3] - p2_array[1, 3]) / w2_const

    # start = time.time_ns()
    x_array = cv2.solve(a_array, b_array, flags=cv2.DECOMP_SVD)
    # x_array, _, _, _ = np.linalg.lstsq(a_array, b_array, rcond=-1) #Alternatively
    # print(f'{ time.time_ns() - start } nsecs')

    return x_array[1]  # for cv2.solve #x_array#for np.linalg.lstsq


def _iter_triangulate_point_w_svd(p1_array,
                                  p2_array,
                                  u1_array,
                                  u2_array):
    """
    Function for Iterative Triangulate Point Using SVD

    (u1_array, p1_array) is the reference pair containing normalized image coordinates (x, y) and the corresponding camera matrix.
    (u2_array, p2_array) is the second pair.

    :param p1_array: [3x4] ndarray
    :param p2_array: [3x4] ndarray
    :param u1_array: [3x1] ndarray
    :param u2_array: [3x1] ndarray

    :return result:
    """

    epsilon = 0.00000000001
    w1_const = 1
    w2_const = 1
    x_array = np.zeros((4, 1), dtype=np.double)
    result = np.zeros((3, 1), dtype=np.double)

    # Hartley suggests 10 iterations at most
    for dummy_idx in range(10):
        x_array_ = _triangulate_point_using_svd(p1_array, p2_array, u1_array, u2_array, w1_const, w2_const)
        x_array[0][0] = x_array_[0][0]
        x_array[1][0] = x_array_[1][0]
        x_array[2][0] = x_array_[2][0]
        x_array[3][0] = 1.0

        p2x1 = p1_array[2, :].dot(x_array)
        p2x2 = p2_array[2, :].dot(x_array)
        # p2x1 and p2x2 should not be zero to avoid RuntimeWarning: invalid value encountered in true_divide
        if (abs(w1_const - p2x1) <= epsilon and abs(w2_const - p2x2) <= epsilon):
            break

        w1_const = p2x1[0]
        w2_const = p2x2[0]

    result[0][0] = x_array[0][0]
    result[1][0] = x_array[1][0]
    result[2][0] = x_array[2][0]

    return result


def new_e2_array(e2_array, r2_array, left_to_right_trans_vector):
    """

    Function to create a new_e2_array which concatenate left_to_right_trans_vector into the last column of e2_array.
    Notes. new_e2_array() is used in triangulate_points_hartley() to avoid too many variables in one method (see R0914)

    :param e2_array: [4x4] narray
    :param r2_array: = left_to_right_rotation_matrix: [3x3] narray
    :param left_to_right_trans_vector: [3x1] narray

    :return e2_array: [4x4] narray
    """
    # While the specification is [3x1] which implies ndarray, the users
    # may also pass in array (3,), ndarray (3, 1) or ndarray (1, 3).
    # So, this will flatten all of them to the same array-like shape.
    l_r = np.ravel(left_to_right_trans_vector)

    for row_idx in range(0, 3):
        for col_idx in range(0, 3):
            e2_array[row_idx, col_idx] = r2_array[row_idx, col_idx]
        e2_array[row_idx, 3] = l_r[row_idx]

    return e2_array


def l2r_to_p2d(p2d, l2r):
    """

    Function to convert l2r array to p2d array, which removes last row of l2r to create p2d.
    Notes. l2r_to_p2d() is used in triangulate_points_hartley() to avoid too many variables in one method (see R0914).

    :param p2d: [3x4] narray
    :param l2r: [4x4] narray

    :return p2d: [3x4] narray
    """

    for dummy_row_index in range(0, 3):
        for dummy_col_index in range(0, 4):
            p2d[dummy_row_index, dummy_col_index] = l2r[dummy_row_index, dummy_col_index]

    return p2d


def triangulate_points_hartley(input_undistorted_points,
                               left_camera_intrinsic_params,
                               right_camera_intrinsic_params,
                               left_to_right_rotation_matrix,
                               left_to_right_trans_vector
                               ):
    """
    Function to compute triangulation of points using Harley

    :param input_undistorted_points: [4x4] narray
    :param left_camera_intrinsic_params: [3x3] narray
    :param right_camera_intrinsic_params: [3x3] narray
    :param left_to_right_rotation_matrix: [3x3] narray
    :param left_to_right_trans_vector: [3x1] narray

    :return outputPoints: [4x3] narray

    References
    ----------
    * Hartley, Richard I., and Peter Sturm. "Triangulation." Computer vision and image understanding 68, no. 2 (1997): 146-157.
    * Prince, Simon JD. Computer vision: models, learning, and inference. Cambridge University Press, 2012.

    :return outputPoints:
    """
    number_of_points = input_undistorted_points.shape[0]
    output_points = np.zeros((number_of_points, 3), dtype=np.double)
    k1_array = left_camera_intrinsic_params
    k2_array = right_camera_intrinsic_params
    _r1_array = np.eye(3, dtype=np.double)  # (unused-variable)
    r2_array = left_to_right_rotation_matrix
    e1_array = np.eye(4, dtype=np.double)
    e2_array = np.eye(4, dtype=np.double)
    p1d = np.zeros((3, 4), dtype=np.double)
    p2d = np.zeros((3, 4), dtype=np.double)

    e2_array = new_e2_array(e2_array, r2_array, left_to_right_trans_vector)
    # Inverting intrinsic params to convert from pixels to normalised image coordinates.
    k1inv = np.linalg.inv(k1_array)
    k2inv = np.linalg.inv(k2_array)

    # Computing coordinates relative to left camera.
    e1inv = np.linalg.inv(e1_array)
    l2r = np.matmul(e2_array, e1inv)

    # The projection matrix, is just the extrinsic parameters, as our coordinates will be in a normalised camera space.
    # P1 should be identity, so that reconstructed coordinates are in Left Camera Space, to P2 should reflect
    # a right to left transform.
    # Prince, Simon JD. Computer vision: models, learning, and inference. Cambridge University Press, 2012.
    p1d[0, 0] = 1
    p1d[0, 1] = 0
    p1d[0, 2] = 0
    p1d[0, 3] = 0
    p1d[1, 0] = 0
    p1d[1, 1] = 1
    p1d[1, 2] = 0
    p1d[1, 3] = 0
    p1d[2, 0] = 0
    p1d[2, 1] = 0
    p1d[2, 2] = 1
    p1d[2, 3] = 0

    p2d = l2r_to_p2d(p2d, l2r)

    u1_array = np.zeros((3, 1), dtype=np.double)
    u2_array = np.zeros((3, 1), dtype=np.double)

    for dummy_index in range(0, number_of_points):
        u1_array[0, 0] = input_undistorted_points[dummy_index, 0]
        u1_array[1, 0] = input_undistorted_points[dummy_index, 1]
        u1_array[2, 0] = 1

        u2_array[0, 0] = input_undistorted_points[dummy_index, 2]
        u2_array[1, 0] = input_undistorted_points[dummy_index, 3]
        u2_array[2, 0] = 1

        # Converting to normalised image points
        u1t = np.matmul(k1inv, u1_array)
        u2t = np.matmul(k2inv, u2_array)

        # array shapes for input args _iter_triangulate_point_w_svd( [3, 4]; [3, 4]; [3, 1]; [3, 1] )
        reconstructed_point = _iter_triangulate_point_w_svd(p1d, p2d, u1t, u2t)

        output_points[dummy_index, 0] = reconstructed_point[0][0]
        output_points[dummy_index, 1] = reconstructed_point[1][0]
        output_points[dummy_index, 2] = reconstructed_point[2][0]

    return output_points


def triangulate_points_opencv(input_undistorted_points,
                              left_camera_intrinsic_params,
                              right_camera_intrinsic_params,
                              left_to_right_rotation_matrix,
                              left_to_right_trans_vector):
    """
    Function to compute triangulation of points using Harley with cv2.triangulatePoints

    :param input_undistorted_points: [4x4] narray

    :param left_camera_intrinsic_params: [3x3] narray
    :param right_camera_intrinsic_params: [3x3] narray
    :param left_to_right_rotation_matrix: [3x3] narray
    :param left_to_right_trans_vector: [3x1] narray

    :return output_points: [4x3] narray

    Other related variables:
        left_undistorted, right_undistorted: point image positions in 2 cameras
        left_undistorted[4x2], right_undistorted[4x2] from input_undistorted_points [4x4]

    References
    ----------
    Hartley, Richard I., and Peter Sturm. "Triangulation." Computer vision and image understanding 68, no. 2 (1997): 146-157.

    """

    l2r_mat = stm.construct_rigid_transformation(left_to_right_rotation_matrix, left_to_right_trans_vector)

    # The projection matrix, is just the extrinsic parameters, as our coordinates will be in a normalised camera space.
    # P1 should be identity, so that reconstructed coordinates are in Left Camera Space, to P2 should reflect
    # a right to left transform.
    # Prince, Simon JD. Computer vision: models, learning, and inference. Cambridge University Press, 2012.
    p1mat = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]], dtype=np.double)

    p2mat = np.zeros((3, 4), dtype=np.double)
    p2mat = l2r_to_p2d(p2mat, l2r_mat)

    number_of_points = input_undistorted_points.shape[0]
    output_points = np.zeros((number_of_points, 3), dtype=np.double)

    # Inverting intrinsic params to convert from pixels to normalised image coordinates.
    k1inv = np.linalg.inv(left_camera_intrinsic_params)
    k2inv = np.linalg.inv(right_camera_intrinsic_params)

    u1_array = np.zeros((3, 1), dtype=np.double)
    u2_array = np.zeros((3, 1), dtype=np.double)

    for dummy_index in range(0, number_of_points):
        u1_array[0, 0] = input_undistorted_points[dummy_index, 0]
        u1_array[1, 0] = input_undistorted_points[dummy_index, 1]
        u1_array[2, 0] = 1

        u2_array[0, 0] = input_undistorted_points[dummy_index, 2]
        u2_array[1, 0] = input_undistorted_points[dummy_index, 3]
        u2_array[2, 0] = 1

        # Converting to normalised image points
        u1t = np.matmul(k1inv, u1_array)
        u2t = np.matmul(k2inv, u2_array)

        # array shapes for input args cv2.triangulatePoints( [3, 4]; [3, 4]; [2, 1]; [2, 1] )
        reconstructed_point = cv2.triangulatePoints(p1mat, p2mat, u1t[:2], u2t[:2])
        reconstructed_point /= reconstructed_point[3][0]  # Homogenize

        output_points[dummy_index, 0] = reconstructed_point[0][0]
        output_points[dummy_index, 1] = reconstructed_point[1][0]
        output_points[dummy_index, 2] = reconstructed_point[2][0]

    return output_points



# -*- coding: utf-8 -*-

""" Various utilities to help testing. """

import os
import glob
import cv2


def load_left_right_pngs(dir_name, expected_number):
    """Given a dir_name, loads left/*.png and right/*.png"""
    left_images = []
    files = glob.glob(os.path.join(dir_name, "left", "*.png"))
    files.sort()
    for file in files:
        image = cv2.imread(file)
        left_images.append(image)
    assert len(left_images) == expected_number

    right_images = []
    files = glob.glob(os.path.join(dir_name, "right", "*.png"))
    files.sort()
    for file in files:
        image = cv2.imread(file)
        right_images.append(image)
    assert len(right_images) == expected_number

    return left_images, right_images



# -*- coding: utf-8 -*-

"""
Tests for calibration_manager.
"""

import glob
import cv2
import pytest
import numpy as np
import sksurgeryimage.calibration.chessboard_point_detector as pd
import sksurgeryimage.calibration.charuco_plus_chessboard_point_detector as chpd
import sksurgerycalibration.video.video_calibration_driver_mono as mcd
import sksurgerycalibration.video.video_calibration_driver_stereo as scd


def load_images_from_glob(glob_pattern):
    """ Load images from files based on a glob pattern. """
    images = []
    files = glob.glob(glob_pattern)
    files.sort()
    for file in files:
        image = cv2.imread(file)
        images.append(image)

    return images


def load_tracking_from_glob(glob_pattern):
    """ Load tracking data from files based on a glob pattern. """
    tracking = []
    files = glob.glob(glob_pattern)
    files.sort()
    for file in files:
        data = np.loadtxt(file)
        tracking.append(data)

    return tracking


def test_handeye_calibration_mono():
    """
    Load mono data (only using left channel) and tracking, do video and
    handeye calibration, compare results against expected values.
    """
    images = load_images_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.left.*.png')

    device_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.device_tracking.*.txt')

    obj_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.calib_obj_tracking.*.txt')

    ref_img = cv2.imread(
        'tests/data/2020_01_20_storz/pattern_4x4_19x26_5_4_with_inset_9x14.png')

    assert len(images) == 10

    assert len(device_tracking) == 10
    assert len(obj_tracking) == 10

    min_number_of_points_per_image = 50
    detector = \
        chpd.CharucoPlusChessboardPointDetector(ref_img,
                                                error_if_no_chessboard=False)

    calibrator = \
        mcd.MonoVideoCalibrationDriver(detector, min_number_of_points_per_image)

    # Grab data from images/tracking arrays
    for image, device, calib_obj in zip(images, device_tracking, obj_tracking):
        successful = calibrator.grab_data(image, device, calib_obj)
        assert successful > 0

    reproj_err_1, _ = calibrator.calibrate()

    assert reproj_err_1 == pytest.approx(1., rel=0.2)

    proj_err, _ = calibrator.handeye_calibration()

    print(f'Reproj err {proj_err}')

    # These values are taken from a previous successful run
    # Not objective measures of correctness
    expected_reproj_error = 10.412839

    assert proj_err == pytest.approx(expected_reproj_error, rel=0.1)


def test_handeye_calibration_stereo():
    """
    Load Stereo data and tracking, do video and
    handeye calibration, compare results against expected values.
    """
    left_images = load_images_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.left.*.png')

    right_images = load_images_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.right.*.png')

    device_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.device_tracking.*.txt')

    obj_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.calib_obj_tracking.*.txt')

    ref_img = cv2.imread(
        'tests/data/2020_01_20_storz/pattern_4x4_19x26_5_4_with_inset_9x14.png')

    assert len(left_images) == 10
    assert len(right_images) == 10
    assert len(device_tracking) == 10
    assert len(obj_tracking) == 10

    min_number_of_points_per_image = 50
    detector = \
        chpd.CharucoPlusChessboardPointDetector(ref_img,
                                                charuco_filtering=True,
                                                error_if_no_chessboard=False)

    calibrator = \
        scd.StereoVideoCalibrationDriver(
            detector, detector, min_number_of_points_per_image)

    # Grab data from images/tracking arrays
    for left, right, device, calib_obj in \
        zip(left_images, right_images, device_tracking, obj_tracking):
        num_left, num_right = calibrator.grab_data(left,
                                                   right,
                                                   device,
                                                   calib_obj)
        assert num_left > 0
        assert num_right > 0
        assert num_left < 480
        assert num_right < 480

    reproj_err_1, recon_err_1, _ = calibrator.calibrate()

    assert reproj_err_1 == pytest.approx(0.6, rel=0.2)
    assert recon_err_1 == pytest.approx(1., rel=0.2)

    proj_err, recon_err, _ = \
        calibrator.handeye_calibration(use_opencv=True, do_bundle_adjust=True)

    print(f'Reproj err {proj_err}')
    print(f'Recon err {recon_err}')

    assert proj_err == pytest.approx(7.3, rel=1.7)
    assert recon_err == pytest.approx(2.5, rel=0.3)

    # test save/load for hand-eye
    calibrator.save_params('tests/output/test_handeye_calibration_stereo', '')
    current_params = calibrator.get_params()

    calibrator.reinit()
    calibrator.load_params('tests/output/test_handeye_calibration_stereo', '')
    loaded_params = calibrator.get_params()
    assert np.allclose(current_params.left_params.handeye_matrix,
                       loaded_params.left_params.handeye_matrix)
    assert np.allclose(current_params.left_params.pattern2marker_matrix,
                       loaded_params.left_params.pattern2marker_matrix)
    assert np.allclose(current_params.right_params.handeye_matrix,
                       loaded_params.right_params.handeye_matrix)
    assert np.allclose(current_params.right_params.pattern2marker_matrix,
                       loaded_params.right_params.pattern2marker_matrix)


def test_load_data_stereo_calib():
    """ Load tracking and image data from test directory. """
    chessboard_detector = pd.ChessboardPointDetector((14, 10), 3, (1, 1))

    stereo_calib = \
        scd.StereoVideoCalibrationDriver(
            chessboard_detector, chessboard_detector, 140)

    tracking_data_dir = 'tests/data/2020_01_20_storz/12_50_30'
    file_prefix = 'calib'

    stereo_calib.load_data(tracking_data_dir, file_prefix)

    assert len(stereo_calib.tracking_data.device_tracking_array) == 10
    assert len(stereo_calib.tracking_data.calibration_tracking_array) == 10
    assert len(stereo_calib.video_data.left_data.images_array) == 10
    assert len(stereo_calib.video_data.right_data.images_array) == 10



# -*- coding: utf-8 -*-

# pylint:disable=line-too-long

""" Various tests to test the SmartLiver calibration using dotty pattern. """

import os
import numpy as np
import cv2
import sksurgerycore.transforms.matrix as skcm
import sksurgerycore.algorithms.procrustes as pbr
import sksurgeryimage.calibration.dotty_grid_point_detector as dgpd
import sksurgerycalibration.video.video_calibration_driver_stereo as sc
import tests.video.test_load_calib_utils as lcu


def get_dotty_calib_driver(calib_dir: str):
    """
    Utility function to setup stereo calibration driver and load data.
    """
    minimum_points_per_frame = 36
    number_of_dots = [18, 25]
    dot_separation = 5
    fiducial_indexes = [133, 141, 308, 316]
    reference_image_size_in_pixels = [1900, 2600]
    pixels_per_mm = 20

    number_of_points = number_of_dots[0] * number_of_dots[1]
    model_points = np.zeros((number_of_points, 6))
    counter = 0
    for y_index in range(number_of_dots[0]):
        for x_index in range(number_of_dots[1]):
            model_points[counter][0] = counter
            model_points[counter][1] = (x_index + 1) * pixels_per_mm
            model_points[counter][2] = (y_index + 1) * pixels_per_mm
            model_points[counter][3] = x_index * dot_separation
            model_points[counter][4] = y_index * dot_separation
            model_points[counter][5] = 0
            counter = counter + 1

    left_intrinsic_matrix = np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.left.intrinsics.txt")
    left_distortion_matrix = np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.left.distortion.txt")
    right_intrinsic_matrix = np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.right.intrinsics.txt")
    right_distortion_matrix = np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.right.distortion.txt")

    threshold_offset = 20
    threshold_window_size = 151

    left_pd = dgpd.DottyGridPointDetector(model_points,
                                          list_of_indexes=fiducial_indexes,
                                          camera_intrinsics=left_intrinsic_matrix,
                                          distortion_coefficients=left_distortion_matrix,
                                          reference_image_size=reference_image_size_in_pixels,
                                          threshold_window_size=threshold_window_size,
                                          threshold_offset=threshold_offset
                                          )

    right_pd = dgpd.DottyGridPointDetector(model_points,
                                           list_of_indexes=fiducial_indexes,
                                           camera_intrinsics=right_intrinsic_matrix,
                                           distortion_coefficients=right_distortion_matrix,
                                           reference_image_size=reference_image_size_in_pixels,
                                           threshold_window_size=threshold_window_size,
                                           threshold_offset=threshold_offset
                                           )

    calibration_driver = sc.StereoVideoCalibrationDriver(left_pd,
                                                         right_pd,
                                                         minimum_points_per_frame)

    for i in range(0,10,2): #issue 59, reduced number of frames used for test
        l_img, r_img, chessboard, scope = lcu.get_calib_data(calib_dir, i)
        num_points = calibration_driver.grab_data(l_img, r_img, scope, chessboard)
        print("Grabbed " + str(num_points) + " points")

    return calibration_driver


def get_pattern_to_marker():
    """
    Creates a pattern 2 marker transform for the SmartLiver calibration pattern.
    """
    plate_points = np.zeros((4, 3))
    plate_points[1][0] = 120
    plate_points[2][1] = 85
    plate_points[3][0] = 120
    plate_points[3][1] = 85

    marker_points = np.ones((4, 3))
    marker_points[0][0] = -21.77
    marker_points[0][2] = -19.25
    marker_points[1][0] = -21.77
    marker_points[1][2] = 100.75
    marker_points[2][0] = -106.77
    marker_points[2][2] = -19.25
    marker_points[3][0] = -106.77
    marker_points[3][2] = 100.75

    p2m_r, p2m_t, _ = pbr.orthogonal_procrustes(marker_points, plate_points)

    pattern2marker = skcm.construct_rigid_transformation(p2m_r, p2m_t)
    return pattern2marker


def get_projection_errs_and_params(dir_name,
                                   pattern2marker,
                                   use_open_cv,
                                   do_bundle_adjust,
                                   save_data=False
                                   ):
    """
    Calibration driver.
    """
    print("dir_name=" + dir_name)

    calib_driver = get_dotty_calib_driver(dir_name)

    reproj_err, recon_err, _ = calib_driver.calibrate()

    print("After initial calibration, reproj_err=" + str(reproj_err) + ", recon_err=" + str(recon_err))

    tracked_reproj_err, tracked_recon_err, params_2 = \
        calib_driver.handeye_calibration(override_pattern2marker=pattern2marker,
                                         use_opencv=use_open_cv,
                                         do_bundle_adjust=do_bundle_adjust
                                         )

    print("After handeye calibration, tracked_reproj_err=" + str(tracked_reproj_err) + ", tracked_recon_err=" + str(tracked_recon_err))
    p2m = params_2.left_params.pattern2marker_matrix
    if p2m is not None:
        print("p2m=" + str(np.transpose(p2m[0:3, 3])))
    print("h2e=" + str(np.transpose(params_2.left_params.handeye_matrix[0:3, 3])))

    if save_data:
        output_calib_dir = os.path.join('tests', 'output', 'test_handeye_dots', dir_name)
        calib_driver.save_data(output_calib_dir, 'calib')
        calib_driver.save_params(output_calib_dir, 'calib')

    return reproj_err, recon_err, tracked_reproj_err, tracked_recon_err, params_2


def get_dirs(is_paper=False):
    """
    Just gets some data directories.
    """
    dirs = []
    if is_paper:
        dirs.append('tests/data/2022_02_11_calibration_viking_paper/18_36_09')
        dirs.append('tests/data/2022_02_11_calibration_viking_paper/18_41_28')
        dirs.append('tests/data/2022_02_11_calibration_viking_paper/18_44_06')
    else:
#        dirs.append('tests/data/2022_02_13_calibration_viking_metal/15_56_22')
#        dirs.append('tests/data/2022_02_13_calibration_viking_metal/15_57_13')
#        dirs.append('tests/data/2022_02_13_calibration_viking_metal/15_58_14')

#        dirs.append('tests/data/2022_02_13_calibration_viking_metal/16_13_39')
#        dirs.append('tests/data/2022_02_13_calibration_viking_metal/16_20_03')
#        dirs.append('tests/data/2022_02_13_calibration_viking_metal/16_24_24')

        dirs.append('tests/data/2022_02_28_calibration_viking_metal/calibration/14_58_31')
        dirs.append('tests/data/2022_02_28_calibration_viking_metal/calibration/15_18_54')
        dirs.append('tests/data/2022_02_28_calibration_viking_metal/calibration/15_22_44')

    return dirs


def get_fixed_pattern_dirs():
    """
    Just gets some data directories.
    """
    dirs = []
    #dirs.append('tests/data/2022_02_28_fixed_position_calibs/calibration/14_58_31')
    dirs.append('tests/data/2022_02_28_fixed_position_calibs/calibration/15_18_54')
#    dirs.append('tests/data/2022_02_28_fixed_position_calibs/calibration/15_22_44')
    return dirs


def get_calibrations(dirs, pattern2marker, use_opencv, use_bundle_adjust, save_data=False):
    """
    Calibration driver.
    """
    results = np.zeros((len(dirs), 8))
    counter = 0
    for dir_name in dirs:
        _, _, tracked_reproj_err, tracked_recon_err, params_2 = \
            get_projection_errs_and_params(dir_name, pattern2marker, use_opencv, use_bundle_adjust, save_data)
        rvec = np.zeros((3,1))
        cv2.Rodrigues(params_2.left_params.handeye_matrix[0:3, 0:3], rvec)
        results[counter][0] = tracked_reproj_err
        results[counter][1] = tracked_recon_err
        results[counter][2] = params_2.left_params.handeye_matrix[0][3]
        results[counter][3] = params_2.left_params.handeye_matrix[1][3]
        results[counter][4] = params_2.left_params.handeye_matrix[2][3]
        results[counter][5] = rvec[0][0]
        results[counter][6] = rvec[1][0]
        results[counter][7] = rvec[2][0]
        counter = counter + 1
    return results


def test_gx_vs_cv():
    """
    Compares consistency of Guofang's method, with OpenCV.
    """
    dirs = get_dirs()
    results_gx = get_calibrations(dirs, None, False, False)
    results_gx_m = np.mean(results_gx, axis=0)
    results_gx_s = np.std(results_gx, axis=0)
    results_opencv = get_calibrations(dirs, None, True, False)
    results_opencv_m = np.mean(results_opencv, axis=0)
    results_opencv_s = np.std(results_opencv, axis=0)
    results_opencv_bundle = get_calibrations(dirs, None, True, True)
    results_opencv_bundle_m = np.mean(results_opencv_bundle, axis=0)
    results_opencv_bundle_s = np.std(results_opencv_bundle, axis=0)

    print("Guofang's mean=\n" + str(results_gx_m))
    print("OpenCV's mean=\n" + str(results_opencv_m))
    print("OpenCV's BA mean=\n" + str(results_opencv_bundle_m))

    print("Guofang's stddev=\n" + str(results_gx_s))
    print("OpenCV's stddev=\n" + str(results_opencv_s))
    print("OpenCV's BA stddev=\n" + str(results_opencv_bundle_s))

    # 1. Just looking at data, not much obvious difference.
    # 2. Interestingly, looking at stddev of rotation and translation of
    #    hand-eye, most difference is translational.
    # 3. Ultimately, chose OpenCV, with no bundle adjustment, as default, as then
    #    we can have more consistent implementation across different use-cases.

    # Just for regression testing:

    # Projection error < 4, should be ok for a stereo laparoscope
    assert results_gx_m[0] < 4
    assert results_opencv_m[0] < 4
    assert results_opencv_bundle_m[0] < 4

    # Reconstruction error < 1mm
    assert results_gx_m[1] < 1
    assert results_opencv_m[1] < 1
    assert results_opencv_bundle_m[1] < 1


def test_fixed_p2m():
    """
    Compares consistency of having a fixed pattern to marker.
    """
    dirs = get_dirs()
    p2m = get_pattern_to_marker()
    results_no_p2m = get_calibrations(dirs, None, True, True)
    results_no_p2m_m = np.mean(results_no_p2m, axis=0)
    results_no_p2m_s = np.std(results_no_p2m, axis=0)
    results_with_p2m = get_calibrations(dirs, p2m, True, True)
    results_with_p2m_m = np.mean(results_with_p2m, axis=0)
    results_with_p2m_s = np.std(results_with_p2m, axis=0)

    print("No p2m mean=\n" + str(results_no_p2m_m))
    print("With p2m mean=\n" + str(results_with_p2m_m))

    print("No p2m stddev=\n" + str(results_no_p2m_s))
    print("With p2m stddev=\n" + str(results_with_p2m_s))

    # Just for regression testing:

    # Projection error < 4, should be ok for a stereo laparoscope
    assert results_no_p2m_m[0] < 4
    assert results_with_p2m_m[0] < 4

    # Reconstruction error < 1mm
    assert results_no_p2m_m[1] < 1
    assert results_with_p2m_m[1] < 1.2


def test_tracked_vs_stationary():
    """
    Mainly to check code runs, as we don't yet have functional requirements.
    Unfortunately, below, the stationary pattern calibrations fail.
    I suspect there was not enough movement of the laparoscope.

    Needs more data. Not a complete unit test, as we aren't
    testing anything other than whether the code runs without failing.
    """
    dirs = get_fixed_pattern_dirs()
    results_fixed = get_calibrations(dirs, None, True, False)
    results_fixed_m = np.mean(results_fixed, axis=0)
    results_fixed_s = np.std(results_fixed, axis=0)
    dirs = get_dirs()
    results_opencv = get_calibrations(dirs, None, True, False)
    results_opencv_m = np.mean(results_opencv, axis=0)
    results_opencv_s = np.std(results_opencv, axis=0)

    print("Fixed pattern mean=\n" + str(results_fixed_m))
    print("OpenCV's mean=\n" + str(results_opencv_m))

    print("Fixed pattern stddev=\n" + str(results_fixed_s))
    print("OpenCV's stddev=\n" + str(results_opencv_s))


def tests_for_smart_liver():
    """
    Additional tests for SmartLiver laparoscope system.
    """
    output_dir = 'tests/output/SmartLiver/'
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    p2m = get_pattern_to_marker()
    np.savetxt(os.path.join(output_dir, 'p2m.txt'), p2m)

    # For documentation, this should re-run calibration, as SmartLiver has it, so should match the input folder.
    results_sl = get_calibrations(
        ['tests/data/2022_02_28_fixed_position_calibs/calibration/14_58_31'],
        pattern2marker=None,
        use_opencv=False,
        use_bundle_adjust=False,
        save_data=False)

    # This should save a new calibration in tests/output,
    # as an example of using a fixed pattern2marker, and OpenCV methods.
    results_2022_03_04 = get_calibrations(
        ['tests/data/2022_02_28_fixed_position_calibs/calibration/14_58_31'],
        pattern2marker=p2m,
        use_opencv=True,
        use_bundle_adjust=False,
        save_data=True)

    print("SmartLiver, proj=" + str(results_sl[0][0]) + ", recon=" + str(results_sl[0][1]))
    print("As of 2022-03-04, proj=" + str(results_2022_03_04[0][0]) + ", recon=" + str(results_2022_03_04[0][1]))



# pylint: disable=line-too-long, missing-module-docstring, invalid-name, missing-function-docstring

import os
from typing import Tuple
import numpy as np
import cv2
import sksurgeryimage.calibration.dotty_grid_point_detector as dotty_pd
import sksurgerycalibration.video.video_calibration_driver_stereo as sc


def get_calib_data(directory: str, idx: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """ Generate filenames for calibration data in a given directory. """
    left_image = cv2.imread(
        os.path.join(directory, f'calib.left.images.{idx}.png')
    )

    right_image = cv2.imread(
        os.path.join(directory, f'calib.right.images.{idx}.png')
    )

    chessboard_tracking = np.loadtxt(
        os.path.join(directory, f'calib.calib_obj_tracking.{idx}.txt')
    )

    scope_tracking = np.loadtxt(
        os.path.join(directory, f'calib.device_tracking.{idx}.txt')
    )

    return left_image, right_image, chessboard_tracking, scope_tracking


def get_point_detector(intrinsic_matrix, distortion_matrix):
    """
    Returns a point detector based on a set of
    camera intrinsics and distortion coefficient parameters.

    :param intrinsic_matrix: [3x3] matrix
    :param distortion_matrix: [1x5] matrix
    :return:
    """
    number_of_dots = [18, 25]
    dot_separation = 3
    fiducial_indexes = [133, 141, 308, 316]
    reference_image_size = [1900, 2600]
    pixels_per_mm = 40

    number_of_points = number_of_dots[0] * number_of_dots[1]
    model_points = np.zeros((number_of_points, 6))
    counter = 0
    for y_index in range(number_of_dots[0]):
        for x_index in range(number_of_dots[1]):
            model_points[counter][0] = counter
            model_points[counter][1] = (x_index + 1) * pixels_per_mm
            model_points[counter][2] = (y_index + 1) * pixels_per_mm
            model_points[counter][3] = x_index * dot_separation
            model_points[counter][4] = y_index * dot_separation
            model_points[counter][5] = 0
            counter = counter + 1

    dot_detector_params = cv2.SimpleBlobDetector_Params()
    dot_detector_params.filterByConvexity = False
    dot_detector_params.filterByInertia = False
    dot_detector_params.filterByCircularity = False
    dot_detector_params.filterByArea = True
    dot_detector_params.minArea = 50
    dot_detector_params.maxArea = 50000

    point_detector = \
        dotty_pd.DottyGridPointDetector(
            model_points,
            fiducial_indexes,
            intrinsic_matrix,
            distortion_matrix,
            reference_image_size=(reference_image_size[1],
                                  reference_image_size[0]),
            threshold_window_size=191,
            dot_detector_params=dot_detector_params
            )

    return point_detector


def get_calib_driver(calib_dir: str):

    left_point_detector = get_point_detector(
        np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.left.intrinsics.txt"),
        np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.left.distortion.txt"))

    right_point_detector = get_point_detector(
        np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.right.intrinsics.txt"),
        np.loadtxt("tests/data/laparoscope_calibration/cbh-viking/calib.right.distortion.txt"))

    minimum_points = 50

    calibration_driver = sc.StereoVideoCalibrationDriver(left_point_detector,
                                                         right_point_detector,
                                                         minimum_points)

    for i in range(3):
        l_img, r_img, chessboard, scope = get_calib_data(calib_dir, i)
        calibration_driver.grab_data(l_img, r_img, scope, chessboard)

    return calibration_driver


def get_ref_dot_detector():

    camera_matrix = np.eye(3)
    distortion_coefficients = np.zeros(5)

    ref_point_detector = get_point_detector(camera_matrix,
                                            distortion_coefficients)
    return ref_point_detector


def test_dotty_grid_iterative():

    os.makedirs("tests/output/iterative", exist_ok=True)

    calib_dir = 'tests/data/dot_calib/11_19_11'

    calib_driver = get_calib_driver(calib_dir)
    ref_detector = get_ref_dot_detector()

    ref_image = cv2.imread('tests/data/dot_calib/circles-25x18-r40-s3.png')
    ref_ids, _, ref_image_points = \
        ref_detector.get_points(ref_image)

    reprojection_err, recon_err, _ = calib_driver.iterative_calibration(
        3,
        ref_ids,
        ref_image_points,
        (ref_image.shape[1],
         ref_image.shape[0]))

    assert reprojection_err < 2
    assert recon_err < 2



# -*- coding: utf-8 -*-

# pylint:disable=invalid-name

"""
Utilities for manipulating quaternions.
"""
import numpy as np


def quat_conjugate(q):
    """
    Obtains the conjugate of a quaternion.
    """
    assert len(q) == 4

    result = [q[0], -q[1], -q[2], -q[3]]

    return np.asarray(result)


def quat_multiply(q, r):
    """
    Calculates the quaternion product n, for two given quaternions q and r:
    n = q x r
    """
    assert len(q) == 4
    assert len(r) == 4
    n0 = r[0] * q[0] - r[1] * q[1] - r[2] * q[2] - r[3] * q[3]
    n1 = r[0] * q[1] + r[1] * q[0] - r[2] * q[3] + r[3] * q[2]
    n2 = r[0] * q[2] + r[1] * q[3] + r[2] * q[0] - r[3] * q[1]
    n3 = r[0] * q[3] - r[1] * q[2] + r[2] * q[1] + r[3] * q[0]

    result = [n0, n1, n2, n3]

    return np.asarray(result)


def quat_to_rotm(q):
    """
    Get the corresponding rotation matrix of a quaternion.
    """
    assert len(q) == 4

    qn = q / np.linalg.norm(q)
    s = qn[0]
    x = qn[1]
    y = qn[2]
    z = qn[3]
    M = [[s**2 + x**2 - y**2 - z**2, 2*(x*y - s*z), 2*(x*z + s*y)],
         [2*(x*y + s*z), s**2 - x**2 + y**2 - z**2, 2*(y*z - s*x)],
         [2*(x*z - s*y), 2*(y*z + s*x), s**2 - x**2 - y**2 + z**2]]

    return np.asarray(M)


def rotm_to_quat_precise(M):
    """
    Get the corresponding quaternion of a rotation matrix.
    Assuming the rotation matrix is orthonormal.
    """
    q = np.empty((4,))
    tr = np.trace(M)
    if tr > 0:
        q[0] = tr + 1.0
        q[3] = M[1, 0] - M[0, 1]
        q[2] = M[0, 2] - M[2, 0]
        q[1] = M[2, 1] - M[1, 2]
    else:
        i, j, k = 0, 1, 2
        if M[1, 1] > M[0, 0]:
            i, j, k = 1, 2, 0
        if M[2, 2] > M[i, i]:
            i, j, k = 2, 0, 1
        tr = M[i, i] - (M[j, j] + M[k, k])
        q[i] = tr + 1.0
        q[j] = M[i, j] + M[j, i]
        q[k] = M[k, i] + M[i, k]
        q[3] = M[k, j] - M[j, k]
        q = q[[3, 0, 1, 2]]

    q *= 0.5 / np.sqrt(tr + 1.0)

    if q[0] < 0.0:
        np.negative(q, q)

    return q


def rotm_to_quat(M):
    """
    Get the corresponding quaternion of a rotation matrix.
    Assuming the rotation matrix is not strictly orthonormal.
    """
    m00 = M[0, 0]
    m01 = M[0, 1]
    m02 = M[0, 2]
    m10 = M[1, 0]
    m11 = M[1, 1]
    m12 = M[1, 2]
    m20 = M[2, 0]
    m21 = M[2, 1]
    m22 = M[2, 2]

    # symmetric matrix K
    K = np.array([[m00-m11-m22, 0.0, 0.0, 0.0],
                  [m01+m10, m11-m00-m22, 0.0, 0.0],
                  [m02+m20, m12+m21, m22-m00-m11, 0.0],
                  [m21-m12, m02-m20, m10-m01, m00+m11+m22]])
    K /= 3.0

    # quaternion is eigenvector of K that corresponds to largest eigenvalue
    w, V = np.linalg.eigh(K)
    q = V[[3, 0, 1, 2], np.argmax(w)]

    if q[0] < 0.0:
        np.negative(q, q)

    return q


def rvec_to_quat(rvec):
    """
    Get the corresponding quaternion of a rotation vector.
    """
    rad = np.linalg.norm(rvec)
    axis = rvec/rad
    q = np.concatenate([[np.cos(rad/2)], np.sin(rad/2) * axis])

    return q


def to_one_hemisphere(quaternions):
    """
    Transform a group of quaternions to one hemisphere.
    """
    new_quaternions = np.zeros(np.shape(quaternions))

    sum_1 = np.sum(np.abs(quaternions[:, 0]))
    sum_2 = np.sum(np.abs(quaternions[:, 1]))
    sum_3 = np.sum(np.abs(quaternions[:, 2]))
    sum_4 = np.sum(np.abs(quaternions[:, 3]))

    max_sum = sum_1
    flag = 0

    if sum_2 > max_sum:
        max_sum = sum_2
        flag = 1

    if sum_3 > max_sum:
        max_sum = sum_3
        flag = 2

    if sum_4 > max_sum:
        flag = 3

    number = np.shape(quaternions)[0]

    for i in range(0, number):
        if quaternions[i, flag] > 0:
            new_quaternions[i] = quaternions[i]
        else:
            new_quaternions[i] = -quaternions[i]

    return new_quaternions



# -*- coding: utf-8 -*-

""" Various routines for Hand-Eye calibration. """

# pylint:disable=invalid-name,consider-using-enumerate

import logging
from typing import List, Tuple

import cv2
import numpy as np
import sksurgerycore.transforms.matrix as mu
from scipy.optimize import least_squares

import sksurgerycalibration.video.quaternion_utils as qu
import sksurgerycalibration.video.video_calibration_utils as vu

LOGGER = logging.getLogger(__name__)


def set_model2hand_arrays(calibration_tracking_array: List,
                          device_tracking_array: List,
                          use_quaternions=False) \
                            -> Tuple[np.ndarray, np.ndarray]:
    """
    Guofang Xiao's method. Set the model-to-hand quaternion
    and translation arrays from tracking data.

    :param calibration_tracking_array: Array of tracking data for
    calibration target
    :type calibration_tracking_array: List of tracking data
    :param device_tracking_array: Array of tracking data for
    device (e.g. camera)
    :type device_tracking_array: List of tracking data
    :param use_quaternions: If True input should be quaternions
    :type device_tracking_array: bool
    :return: quaternion model to hand array and translation model to hand
    array
    :rtype: np.ndarray, np.ndarray
    """

    if len(calibration_tracking_array) != len(device_tracking_array):
        raise ValueError('Calibration target and device tracking array \
                            should be the same size')

    number_of_frames = len(calibration_tracking_array)

    quat_model2hand_array = np.zeros((number_of_frames, 4))
    trans_model2hand_array = np.zeros((number_of_frames, 3))

    for i in range(number_of_frames):
        if use_quaternions:
            quat_model = \
                calibration_tracking_array[i][0, 0:4]
            trans_model = \
                calibration_tracking_array[i][0, 4:7]
            quat_hand = \
                device_tracking_array[i][0, 0:4]
            trans_hand = \
                device_tracking_array[i][0, 4:7]

            quat_model2hand = qu.quat_multiply(qu.quat_conjugate(quat_hand),
                                            quat_model)
            q_translation = np.append([0], (trans_model - trans_hand))
            quat_hand_conj = qu.quat_conjugate(quat_hand)

            trans_model2hand = qu.quat_multiply(quat_hand_conj,
                                                qu.quat_multiply(q_translation,
                                                                 quat_hand))
            trans_model2hand = trans_model2hand[1:4]
        else:
            tracking_model = calibration_tracking_array[i]
            tracking_hand = device_tracking_array[i]
            model_to_hand = np.linalg.inv(tracking_hand) @ tracking_model

            quat_model2hand = qu.rotm_to_quat(model_to_hand[0:3, 0:3])
            trans_model2hand = model_to_hand[0:3, 3]

        quat_model2hand_array[i] = quat_model2hand
        trans_model2hand_array[i] = trans_model2hand

    quat_model2hand_array = \
        qu.to_one_hemisphere(quat_model2hand_array)

    return quat_model2hand_array, trans_model2hand_array


def _gx_solve_2quaternions(qx, q_A, q_B):
    """
    Guofang Xiao's method. Provide cost function for least-square optimisation
    to solve quaternions qx1 and qx2 in: q_B = qx1 * q_A * qx2

    :param qx: 1x8 vector of 2 quaternions
    :param q_A: Nx4 matrices of N quaternions
    :param q_B: Nx4 matrices of N quaternions
    :return:
    """
    if np.shape(q_A) != np.shape(q_B):
        raise AssertionError("q_A and q_B must have the same shape.")

    number_of_frames = np.shape(q_A)[0]

    f = []

    qx_1 = qx[0:4]
    qx_2 = qx[4:8]

    qx_1 = qx_1/np.linalg.norm(qx_1)
    qx_2 = qx_2/np.linalg.norm(qx_2)

    for i in range(0, number_of_frames):
        f.append(qu.quat_multiply(qu.quat_multiply(qx_1, q_A[i]), qx_2)
                 - q_B[i])

    f = np.ndarray.flatten(np.asarray(f))

    return f


def _gx_solve_2translations(q_handeye,
                            quat_model2hand_array,
                            trans_model2hand_array,
                            trans_extrinsics_array):
    """
    Guofang Xiao's method. Solve translations t_handeye and t_pattern2marker.
    translations = [t_handeye t_pattern2marker]

    :param q_handeye: 1x4 matrix quaternion
    :param quat_model2hand_array: Nx4 matrix of N quaternions
    :param trans_model2hand_array: Nx3 matrices of N translations
    :param trans_extrinsics_array: Nx3 matrices of extrinsic translation vectors
    :return:
    """
    if np.shape(quat_model2hand_array)[0] \
            != np.shape(trans_model2hand_array)[0]:
        raise AssertionError("Wrong dimensions in solve_2translations().")
    if np.shape(trans_model2hand_array) != np.shape(trans_extrinsics_array):
        raise AssertionError("Wrong dimensions in solve_2translations().")

    number_of_frames = np.shape(quat_model2hand_array)[0]

    A = np.zeros((3 * number_of_frames, 6))
    B = np.zeros((3 * number_of_frames, 1))

    for i in range(0, number_of_frames):
        qHEqMH = qu.quat_multiply(q_handeye, quat_model2hand_array[i])
        rHErMH = qu.quat_to_rotm(qHEqMH)
        rHE = qu.quat_to_rotm(q_handeye)

        A[i*3:i*3+3, :] = np.hstack([np.identity(3), rHErMH])
        b_i = np.transpose(trans_extrinsics_array[i, :])\
            - np.matmul(rHE, np.transpose(trans_model2hand_array[i, :]))
        B[i*3:i*3+3] = np.reshape(b_i, (3, 1))

    A_T = np.transpose(A)
    translations = np.linalg.inv(A_T @ A) @ A_T @ B
    translations = np.transpose(translations)

    return translations.flatten()


def _gx_handeye_optimisation(
        quat_extrinsics_array: np.ndarray,
        trans_extrinsics_array: np.ndarray,
        quat_model2hand_array: np.ndarray,
        trans_model2hand_array: np.ndarray) ->  \
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Guofang Xiao's method. Solve handeye and pattern-to-marker transformations.

    :param quat_extrinsics_array: An array of quaternions representing the
        rotations of the camera extrinsics.
    :type quat_extrinsics_array: np.ndarray
    :param trans_extrinsics_array:  Array of the translation vectors of the
        camera extrinsics.
    :type trans_extrinsics_array: np.ndarray
    :param quat_model2hand_array: An array of model to hand quaternions.
        :type quat_model2hand_array: np.ndarray
    :param quat_model2hand_array: An array of model to hand
        translations arrays.
        :type quat_model2hand_array: np.ndarray
    :return: rotations in quaternion form and translations of the handeye
        pattern-to-marker transformation.
    :rtype: np.ndarray
    """

    # self.quat_model2hand_array should already be in one hemisphere.

    quat_extrinsics_array = qu.to_one_hemisphere(quat_extrinsics_array)

    qx_0 = [1, 0, 0, 0, 1, 0, 0, 0]
    lb = [-1, -1, -1, -1, -1, -1, -1, -1]
    ub = [1, 1, 1, 1, 1, 1, 1, 1]

    op_result = least_squares(_gx_solve_2quaternions, qx_0,
                              bounds=(lb, ub),
                              args=(quat_model2hand_array,
                                    quat_extrinsics_array)
                              )

    q_handeye = op_result.x[0:4]
    q_pattern2marker = op_result.x[4:8]

    q_handeye = q_handeye / np.linalg.norm(q_handeye)
    q_pattern2marker = q_pattern2marker / np.linalg.norm(q_pattern2marker)

    translations = _gx_solve_2translations(q_handeye,
                                           quat_model2hand_array,
                                           trans_model2hand_array,
                                           trans_extrinsics_array)

    t_handeye = translations[0:3]
    t_pattern2marker = translations[3:6]

    return q_handeye, t_handeye, q_pattern2marker, t_pattern2marker


def guofang_xiao_handeye_calibration(rvecs: List[np.ndarray],
                                     tvecs: List[np.ndarray],
                                     quat_model2hand_array: np.ndarray,
                                     trans_model2hand_array: np.ndarray) \
                                     -> Tuple[np.ndarray, np.ndarray]:
    """
    Guofang Xiao's method. Solve for the hand-eye transformation,
    as well as the transformation from the pattern to the tracking markers
    on the calibration object.

    This method, developed for the SmartLiver project, assumes both
    device (laparoscope), and the calibration object are tracked. We also, keep
    the device (laparoscope) stationary while moving the calibration object.

    :param rvecs: Array of rotation vectors, from OpenCV, camera extrinsics
    (pattern to camera transform)
    :type rvecs: List[np.ndarray]
    :param tvecs: Array of translation vectors, from OpenCV, camera extrinsics
    (pattern to camera transform)
    :type tvecs: List[np.ndarray]
    :param quat_model2hand_array: Array of quaternions representing rotational
    part of marker-to-hand.
    :type quat_model2hand_array: np.ndarray
    :param trans_model2hand_array: Array of quaternions representing
    translational part of marker-to-hand.
    :type trans_model2hand_array: np.ndarray
    :return: two 4x4 matrices as np.ndarray, representing handeye_matrix,
    pattern2marker_matrix
    :rtype: np.ndarray, np.ndarray
    """

    number_of_frames = len(rvecs)
    quat_extrinsics_array = np.zeros((number_of_frames, 4))

    for i in range(number_of_frames):
        quat_extrinsics_array[i] = qu.rvec_to_quat(rvecs[i].flatten())

    trans_extrinsics_array = np.reshape(tvecs, (number_of_frames, 3))

    q_handeye, t_handeye, q_pattern2marker, t_pattern2marker = \
        _gx_handeye_optimisation(quat_extrinsics_array, trans_extrinsics_array,
                                 quat_model2hand_array, trans_model2hand_array)

    handeye_matrix = np.eye(4)
    handeye_matrix[0:3, 0:3] = qu.quat_to_rotm(q_handeye)
    handeye_matrix[0:3, 3] = t_handeye

    pattern2marker_matrix = np.eye(4)
    pattern2marker_matrix[0:3, 0:3] = qu.quat_to_rotm(q_pattern2marker)
    pattern2marker_matrix[0:3, 3] = t_pattern2marker

    return handeye_matrix, pattern2marker_matrix


def calibrate_hand_eye_using_stationary_pattern(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_HAND_EYE_TSAI,
        invert_camera=False
        ):
    """
    Hand-eye calibration using standard OpenCV methods. This method assumes
    a single set of tracking data, so it is useful for a stationary, untracked
    calibration pattern, and a tracked video device, e.g. laparoscope.

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param tracking_matrices: list of tracking matrices for the tracked
    device, marker_to_tracker.
    :param method: Choice of OpenCV Hand-Eye method.
    :param invert_camera: if True, we invert camera matrices before
    hand-eye calibration.
    :return hand-eye transform as 4x4 matrix.
    """
    if len(camera_rvecs) != len(camera_tvecs):
        raise ValueError("Camera rotation and translation vector "
                         "lists must be the same length.")

    if len(camera_tvecs) != len(tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms must "
                         "equal the number of tracking matrices.")

    if len(camera_rvecs) < 3:
        raise ValueError("You must have at least 3 views, include movements "
                         "around at least 2 different rotation axes.")

    # Convert tracking matrices to rvecs/tvecs for OpenCV.
    tracking_rvecs = []
    tracking_tvecs = []
    for i in range(len(camera_rvecs)):
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(tracking_matrices[i])
        tracking_rvecs.append(rvecs)
        tracking_tvecs.append(tvecs)

    cam_rvecs = []
    cam_tvecs = []
    for i in range(len(camera_rvecs)):
        mat = vu.extrinsic_vecs_to_matrix(camera_rvecs[i], camera_tvecs[i])
        if invert_camera:
            mat = np.linalg.inv(mat)
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(mat)
        cam_rvecs.append(rvecs)
        cam_tvecs.append(tvecs)

    # OpenCV outputs eye-to-hand.
    e2h_rmat, e2h_tvec = cv2.calibrateHandEye(tracking_rvecs,
                                              tracking_tvecs,
                                              cam_rvecs,
                                              cam_tvecs,
                                              method=method
                                              )

    e2h = mu.construct_rigid_transformation(e2h_rmat, e2h_tvec)
    h2e = np.linalg.inv(e2h)

    return h2e


def calibrate_hand_eye_using_tracked_pattern(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        device_tracking_matrices: List[np.ndarray],
        pattern_tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_HAND_EYE_TSAI
        ):
    """
    Hand-eye calibration using standard OpenCV methods. This method assumes
    two sets of tracking data, so it is useful for a tracked device
    (e.g. laparoscope) and a tracked calibration object (e.g. chessboard).

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param device_tracking_matrices: list of tracking matrices for the
    tracked device, marker_to_tracker.
    :param pattern_tracking_matrices: list of tracking matrices for the
    tracked calibration object, marker_to_tracker.
    :param method: Choice of OpenCV Hand-Eye method.
    :return hand-eye transform as 4x4 matrix.
    """
    if len(device_tracking_matrices) != len(pattern_tracking_matrices):
        raise ValueError("The number of device tracking matrices must "
                         "equal the number of pattern tracking matrices.")

    if len(device_tracking_matrices) < 3:
        raise ValueError("You must have at least 3 tracking matrices")

    for i in range(len(pattern_tracking_matrices)):
        if pattern_tracking_matrices[i] is None:
            raise ValueError("Pattern tracking matrix:"
                             + str(i) + str(" is None."))

    # Essentially, as we have a tracked calibration object,
    # the calibration object becomes the reference for the tracker space.
    tracking_matrices = []
    for i in range(len(device_tracking_matrices)):
        mat = np.linalg.inv(pattern_tracking_matrices[i]) \
              @ device_tracking_matrices[i]
        tracking_matrices.append(mat)

    h2e = calibrate_hand_eye_using_stationary_pattern(camera_rvecs,
                                                      camera_tvecs,
                                                      tracking_matrices,
                                                      method)
    return h2e


def calibrate_pattern_to_tracking_marker(camera_rvecs: List[np.ndarray],
                                         camera_tvecs: List[np.ndarray],
                                         tracking_matrices: List[np.ndarray],
                                         method=cv2.CALIB_HAND_EYE_TSAI
                                         ):
    """
    In some calibration problems, people use a pattern (e.g. chessboard)
    that is tracked. If you manufacture this well, you should know the
    pattern_2_marker transform by DESIGN. However, if you assume a STATIONARY
    video camera, and a moving pattern, this method allows you to use standard
    hand-eye techniques to estimate the pattern_2_marker transform from
    at least 2 motions (3 views) around different axes of rotation.

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param tracking_matrices: list of tracking matrices for the tracked
    calibration pattern, marker_to_tracker.
    :param method: Choice of OpenCV Hand-Eye method.
    :return pattern_to_marker
    """
    h2e = calibrate_hand_eye_using_stationary_pattern(camera_rvecs,
                                                      camera_tvecs,
                                                      tracking_matrices,
                                                      method,
                                                      invert_camera=True)

    p2m = np.linalg.inv(h2e)

    return p2m


def calibrate_hand_eye_and_pattern_to_marker(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        device_tracking_matrices: List[np.ndarray],
        pattern_tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_ROBOT_WORLD_HAND_EYE_SHAH
        ):
    """
    Hand-eye calibration using standard OpenCV methods. This method assumes
    you are tracking both the device that needs hand-eye calibration, and
    the calibration pattern.

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param device_tracking_matrices: list of tracking matrices for the
    tracked device, e.g. laparoscope, marker_to_tracker.
    :param pattern_tracking_matrices: list of tracking matrices for the
    calibration object, marker_to_tracker
    :param method: Choice of OpenCV RobotWorldHandEye method.
    :return hand-eye, pattern-to-marker transforms as 4x4 matrices
    """
    if len(camera_rvecs) != len(camera_tvecs):
        raise ValueError("Camera rotation and translation vector lists "
                         "must be the same length.")

    if len(camera_tvecs) != len(device_tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms must "
                         "equal the number of device tracking matrices.")

    if len(camera_tvecs) != len(pattern_tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms must "
                         "equal the number of pattern tracking matrices.")

    if len(camera_rvecs) < 3:
        raise ValueError("You must have at least 3 views, include movements "
                         "around at least 2 different rotation axes.")

    # Convert tracking matrices to rvecs/tvecs for OpenCV.
    # Look at OpenCV documentation for cv.calibrationRobotWorldHandEye
    # Gripper = tracking marker on calibration pattern
    # Robot base = tracking marker on video device, eg. laparoscope
    # World = video camera on laparoscope, assumed to be stationary
    # So, we move the calibration pattern, not the video device.

    base2gripper_rvecs = []
    base2gripper_tvecs = []
    for i in range(len(camera_rvecs)):
        b2g = np.linalg.inv(pattern_tracking_matrices[i]) \
              @ device_tracking_matrices[i]
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(b2g)
        base2gripper_rvecs.append(rvecs)
        base2gripper_tvecs.append(tvecs)

    cam_rvecs = []
    cam_tvecs = []
    for i in range(len(camera_rvecs)):
        mat = vu.extrinsic_vecs_to_matrix(camera_rvecs[i], camera_tvecs[i])
        mat = np.linalg.inv(mat)
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(mat)
        cam_rvecs.append(rvecs)
        cam_tvecs.append(tvecs)

    b2w_rmat, b2w_t, g2c_rmat, g2c_t = \
        cv2.calibrateRobotWorldHandEye(cam_rvecs,
                                       cam_tvecs,
                                       base2gripper_rvecs,
                                       base2gripper_tvecs,
                                       method=method)

    h2e = mu.construct_rigid_transformation(b2w_rmat, b2w_t)
    m2p = mu.construct_rigid_transformation(g2c_rmat, g2c_t)
    p2m = np.linalg.inv(m2p)

    return h2e, p2m


def calibrate_hand_eye_and_grid_to_world(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        device_tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_ROBOT_WORLD_HAND_EYE_SHAH
        ):
    """
    Hand-eye calibration using standard OpenCV methods. This method assumes
    you have a stationary untracked calibration pattern, and a tracked
    device (e.g. laparoscope)

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param device_tracking_matrices: list of tracking matrices for the
    tracked device, e.g. laparoscope, marker_to_tracker.
    :param method: Choice of OpenCV RobotWorldHandEye method.
    :return hand-eye, grid-to-world transforms as 4x4 matrices
    """
    if len(camera_rvecs) != len(camera_tvecs):
        raise ValueError("Camera rotation and translation vector "
                         "lists must be the same length.")

    if len(camera_tvecs) != len(device_tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms "
                         "must equal the number of device tracking matrices.")

    if len(camera_rvecs) < 3:
        raise ValueError("You must have at least 3 views, include "
                         "movements around at least 2 different rotation axes.")

    base2gripper_rvecs = []
    base2gripper_tvecs = []
    for i in range(len(device_tracking_matrices)):
        b2g = np.linalg.inv(device_tracking_matrices[i])
        rvec, tvec = vu.extrinsic_matrix_to_vecs(b2g)
        base2gripper_rvecs.append(rvec)
        base2gripper_tvecs.append(tvec)

    b2w_rmat, b2w_t, g2c_rmat, g2c_t = \
        cv2.calibrateRobotWorldHandEye(camera_rvecs,
                                       camera_tvecs,
                                       base2gripper_rvecs,
                                       base2gripper_tvecs,
                                       method=method)

    h2e = mu.construct_rigid_transformation(g2c_rmat, g2c_t)
    b2w = mu.construct_rigid_transformation(b2w_rmat, b2w_t)
    w2b = np.linalg.inv(b2w)

    return h2e, w2b



# pylint: disable=line-too-long, missing-module-docstring, invalid-name, missing-function-docstring

import numpy as np
import cv2
import sksurgeryimage.calibration.charuco_plus_chessboard_point_detector \
    as charuco_pd
import sksurgerycalibration.video.video_calibration_driver_stereo as sc
import tests.video.test_load_calib_utils as lcu


def get_calib_driver(calib_dir: str):
    """ Create left/right charuco point detectors and load calibration images from directory. """
    reference_image = cv2.imread("tests/data/2020_01_20_storz/pattern_4x4_19x26_5_4_with_inset_9x14.png")
    minimum_points = 50

    number_of_squares = [19, 26]
    square_tag_sizes = [5, 4]
    filter_markers = True
    number_of_chessboard_squares = [9, 14]
    chessboard_square_size = 3
    chessboard_id_offset = 500

    left_pd = \
        charuco_pd.CharucoPlusChessboardPointDetector(
            reference_image,
            minimum_number_of_points=minimum_points,
            number_of_charuco_squares=number_of_squares,
            size_of_charuco_squares=square_tag_sizes,
            charuco_filtering=filter_markers,
            number_of_chessboard_squares=number_of_chessboard_squares,
            chessboard_square_size=chessboard_square_size,
            chessboard_id_offset=chessboard_id_offset
        )

    right_pd = \
        charuco_pd.CharucoPlusChessboardPointDetector(
            reference_image,
            minimum_number_of_points=minimum_points,
            number_of_charuco_squares=number_of_squares,
            size_of_charuco_squares=square_tag_sizes,
            charuco_filtering=filter_markers,
            number_of_chessboard_squares=number_of_chessboard_squares,
            chessboard_square_size=chessboard_square_size,
            chessboard_id_offset=chessboard_id_offset
        )

    calibration_driver = sc.StereoVideoCalibrationDriver(left_pd,
                                                         right_pd,
                                                         minimum_points)

    for i in range(3):
        l_img, r_img, chessboard, scope = lcu.get_calib_data(calib_dir, i)
        calibration_driver.grab_data(l_img, r_img, scope, chessboard)

    return calibration_driver


# Two datasets, A and B.
# Independently calibrating them gives Stereo reprojection error < 1
# But if we pass the intrinsics from A as precalibration for B, then
# error is ~4, so potentially something fishy going on.
def test_charuco_dataset_A():

    calib_dir = 'tests/data/precalib/precalib_base_data'
    calib_driver = get_calib_driver(calib_dir)

    stereo_reproj_err, stereo_recon_err, _ = \
        calib_driver.calibrate()

    tracked_reproj_err, tracked_recon_err, _ = \
        calib_driver.handeye_calibration()

    print(stereo_reproj_err, stereo_recon_err, tracked_reproj_err, tracked_recon_err)
    assert stereo_reproj_err < 1
    assert stereo_recon_err < 4
    assert tracked_reproj_err < 3
    assert tracked_recon_err < 4


def test_charuco_dataset_B():

    calib_dir = 'tests/data/precalib/data_moved_scope'
    calib_driver = get_calib_driver(calib_dir)

    stereo_reproj_err, stereo_recon_err, _ = \
        calib_driver.calibrate()

    tracked_reproj_err, tracked_recon_err, _ = \
        calib_driver.handeye_calibration()

    print(stereo_reproj_err, stereo_recon_err, tracked_reproj_err, tracked_recon_err)
    assert stereo_reproj_err < 1
    assert stereo_recon_err < 3
    assert tracked_reproj_err < 4
    assert tracked_recon_err < 3


def test_precalbration():
    """ Use intrinsics from A to calibration B, currently failing. """
    left_intrinsics = np.loadtxt('tests/data/precalib/precalib_base_data/calib.left.intrinsics.txt')
    left_distortion = np.loadtxt('tests/data/precalib/precalib_base_data/calib.left.distortion.txt')
    right_intrinsics = np.loadtxt('tests/data/precalib/precalib_base_data/calib.right.intrinsics.txt')
    right_distortion = np.loadtxt('tests/data/precalib/precalib_base_data/calib.right.distortion.txt')
    l2r = np.loadtxt('tests/data/precalib/precalib_base_data/calib.l2r.txt')
    l2r_rmat = l2r[0:3, 0:3]
    l2r_tvec = l2r[0:3, 3]

    calib_dir = 'tests/data/precalib/data_moved_scope'
    calib_driver = get_calib_driver(calib_dir)

    stereo_reproj_err, stereo_recon_err, _ = \
        calib_driver.calibrate(
            override_left_intrinsics=left_intrinsics,
            override_left_distortion=left_distortion,
            override_right_intrinsics=right_intrinsics,
            override_right_distortion=right_distortion,
            override_l2r_rmat=l2r_rmat,
            override_l2r_tvec=l2r_tvec)

    tracked_reproj_err, tracked_recon_err, _ = \
        calib_driver.handeye_calibration()

    print(stereo_reproj_err, stereo_recon_err, tracked_reproj_err, tracked_recon_err)
    assert stereo_reproj_err < 4.5
    assert stereo_recon_err < 4.5
    assert tracked_reproj_err < 5.4
    assert tracked_recon_err < 6.4
